unit PreProcessorU;
// DONE 5 : Improve pre-processor speed
{$DEFINE ConvertCommentsToWhiteSpace}
{.$DEFINE BreakOnError}
{$DEFINE CaseSensitive}
interface
uses
  sysutils,
  classes,
  ScannerU;

Type
  EPreProcessorError = class(Exception);

  TPreProcessor = class;

  TPreProcessorCommandEvent = Procedure ( sender : TPreProcessor ) of object;
  TExpandMacroEvent = function (const MacroName : string; var NewText : string) : boolean of object;

  TPreProcessor = Class
  Private
    ftext : string;
    // the index into the source text of the next char
    fNextCharIndex : integer;
    // look ahead char
    fLook,fNextLook : char;
    fToken : TTokenData;
    fOnPreProcessorCommand : TPreProcessorCommandEvent;
    fOnExpandMacro : TExpandMacroEvent;
    fDoExpandMacro : boolean;

    // error methods
    procedure Error(Const s: string; ErrorAddress: Pointer);

    procedure Expected(Const s,a: string); overload;
    procedure Expected(Const s,a: string; ErrorAddress: Pointer); overload;
    Procedure Scanner_Abort(Const s : string);

    Procedure SetNextCharIndex(value : integer);
    Procedure SetText(const aText : string);
  public
    // Read New Character From Input Stream
    function GetChar : char;
    // Skip Over a commented section, if 'ConvertCommentsToWhiteSpace' is defined,
    // the comments are converted into a white space char
    Procedure SkipComment;
    // Skip Over Leading White Space
    procedure SkipWhite;

    // Get an Identifier
    Procedure GetIdentifier;
    // Get a string
    procedure GetString;    
    // Get a Number
    Procedure GetNumber;
    Procedure GetOperator;
  Public
    Constructor Create( Const atext : string;
                        aOnPreProcessorCommand : TPreProcessorCommandEvent;
                        aOnExpandMacro : TExpandMacroEvent);
    Destructor Destroy; override;

    // Skip Over a Comma if it exists.
    function SkipComma : boolean;

    // loads a token into 'token', and its value into 'value'
    // the value reported is a step behind the scanner, this allows
    // the next token (the current token for the scanner) to be access able to the
    // Parser
    // 'DelayPreProcessor' causes it to flush the token & Next Token, and only invoke the
    // preprocessor AFTER it has done this
    Procedure GetToken;  // Lexical Scanner

    // Skips the the next new line
    procedure SkipToNewLine;
    Procedure SkipNewLine;
    // returns the string of characters before the next newline
    Function TillNewLine( DoSkipComments : boolean = true ) : string;
    // Inserts a character into the text stream
    Procedure InsertString( index : integer;
                            const substr : string;
                            AutoSkipWhiteSpace: boolean = true );
    Procedure RemoveString( index, count : integer;
                            AutoSkipWhiteSpace: boolean = true );

    Procedure Match( aToken : TToken );
    Function IfMatch( aToken : TToken ): boolean;

    Property CurrentTokenData : TTokenData read fToken write fToken;

    Property Token : TToken read fToken.Token write fToken.Token;
    Property Value : string read fToken.Value write fToken.Value;
    Property Position : integer read fToken.Position write fToken.Position;

    Property NextCharIndex : integer read fNextCharIndex write SetNextCharIndex;
    Property Look : char read fLook write fLook;
    Property NextLook : char read fNextLook write fNextLook;

    Property OnPreProcessorCommand : TPreProcessorCommandEvent read fOnPreProcessorCommand write fOnPreProcessorCommand;
    Property OnExpandMacro : TExpandMacroEvent read fOnExpandMacro write fOnExpandMacro;
    Property DoExpandMacro : boolean read fDoExpandMacro write fDoExpandMacro;

    Property Text : string read ftext write SetText;
  end; {TPreProcessor}

implementation

// -----------------------------------------------------------------------------

Constructor TPreProcessor.Create( const atext : string;
                                  aOnPreProcessorCommand : TPreProcessorCommandEvent;
                                  aOnExpandMacro : TExpandMacroEvent);
begin
fOnPreProcessorCommand := aOnPreProcessorCommand;
fOnExpandMacro := aOnExpandMacro;
fDoExpandMacro := true;
Text := atext;
end; {Create}

Procedure TPreProcessor.SetText(const aText : string);
begin
fText := aText;
// set up the input stream
NextCharIndex := 2;
GetToken; // pre-load token
end; {SetText}


Destructor TPreProcessor.Destroy;
begin
Inherited Destroy;
end; {destroy}

Procedure TPreProcessor.Error(Const s: string; ErrorAddress: Pointer);
begin
//WriteMsg('');
//WriteMsg('Error: '+ s+ '.');
{$IFDEF BreakOnError}
asm int 3 end;
{$ENDIF}
raise EPreProcessorError.create('Error: '+ s+ '.') at ErrorAddress;
end; {Error}

procedure TPreProcessor.Expected(Const s,a: string);
var ErrorAddress : Pointer;
begin
asm
  MOV eax,[EBP+4];
  MOV ErrorAddress,eax;
end;
Error('"'+s + '" expected but "' +a+ '" found',ErrorAddress);
end; {Expected}

procedure TPreProcessor.Expected(Const s,a: string; ErrorAddress: Pointer);
begin
Error('"'+s + '" expected but "' +a+ '" found',ErrorAddress);
end; {Expected}

Procedure TPreProcessor.Scanner_Abort(Const s : string);
var ErrorAddress : Pointer;
begin
// load the error data
fToken.Value := look;
fToken.Token := CSTINT_WhiteSpace;
fToken.Position := NextCharIndex-1;
// Report the error
asm
  MOV eax,[EBP+4];
  MOV ErrorAddress,eax;
end;
Error(s, ErrorAddress);
end; {Scanner_Abort}

// -----------------------------------------------------------------------------

// Read New Character From Input Stream
function TPreProcessor.GetChar : char;
begin
look := NextLook;
if NextCharIndex < length(ftext) then
  begin
  inc(fNextCharIndex);
  NextLook := ftext[NextCharIndex];
  end
else
  NextLook := #0;
result := look;
end; {GetChar}

Procedure TPreProcessor.SetNextCharIndex(value : integer);
begin
fNextCharIndex := value;
if ( NextCharIndex = 0) or
   ( NextCharIndex = 1) and
   ( length(ftext) >= 1) then
  begin
  fNextCharIndex := 2;
  Look := ftext[1];
  if length(ftext) >= 2 then
    NextLook := ftext[2];
  end
else
  begin
  if (NextCharIndex > 0) and
     (NextCharIndex-1 < length(ftext)) then
    Look := ftext[NextCharIndex-1]
  else
    Look := #0;
  if (NextCharIndex > 0) and
     (NextCharIndex < length(ftext)) then
    NextLook := ftext[NextCharIndex]
  else
    NextLook := #0;
  end;
end; {SetNextCharIndex}

// Inserts a character into the text stream
Procedure TPreProcessor.InsertString( index : integer;
                                 const substr : string;
                                 AutoSkipWhiteSpace: boolean = true );
begin
insert(substr, ftext, index);
NextCharIndex := NextCharIndex;
if Position > index then
  Position := Position + length(substr);
if AutoSkipWhiteSpace then
  SkipWhite;
end; {InsertString}

Procedure TPreProcessor.RemoveString( index,count : integer;
                                      AutoSkipWhiteSpace: boolean = true);
begin
if index < 1 then
  index := 1;
Delete(fText,index,count);
if NextCharIndex >= index then
  NextCharIndex := NextCharIndex - count;
if NextCharIndex > length(fText) then
  NextCharIndex := length(fText);
if Position > index then
  Position := Position - index;
if AutoSkipWhiteSpace then
  SkipWhite;
end; {RemoveString}

// -----------------------------------------------------------------------------

// Skip Over a commented section, if 'ConvertCommentsToWhiteSpace' is defined,
// the comments are converted into a white space char
Procedure TPreProcessor.SkipComment;
var
  count : Longword;
  StartIndex : longint;
{$IFDEF ConvertCommentsToWhiteSpace}
  SkippedComment : boolean;
{$ENDIF}
begin
{$IFDEF ConvertCommentsToWhiteSpace}
SkippedComment := false;
{$ENDIF}
StartIndex := NextCharIndex -1;
// comment type one
if (Look = CommentType1[1]) and (NextLook = CommentType1[2]) then
  begin
  SkipToNewLine;
  RemoveString( StartIndex, NextCharIndex - StartIndex-1, false );
//  SkipNewLine;  
{$IFDEF ConvertCommentsToWhiteSpace}
  SkippedComment := true;
{$ENDIF}
  end;
// comment type 2
if (Look = CommentType2Start[1]) and (NextLook =CommentType2Start[2]) then
  begin
  count := 0;
  repeat
    if (Look = CommentType2Start[1]) and (NextLook =CommentType2Start[2]) then
      begin
      inc(count);
      GetChar;
      end
    else if (Look = CommentType2End[1]) and (NextLook =CommentType2End[2]) then
      begin
      dec(count);
      GetChar;
      end;
    GetChar;
    if look = #0 then
      Scanner_Abort('Unexpected end of file in comment');
  Until count = 0;
  RemoveString( StartIndex, NextCharIndex - StartIndex-1, false );
{$IFDEF ConvertCommentsToWhiteSpace}
  SkippedComment := true;
{$ENDIF}
  end;
{$IFDEF ConvertCommentsToWhiteSpace}
if SkippedComment then
  begin
  Look := ' ';
  dec( fNextCharIndex );
  if fNextCharIndex < 2 then
    fNextCharIndex := 2;
  if fNextCharIndex < Length( ftext ) then
    NextLook := ftext[ fNextCharIndex ]
  else
    NextLook := #0;
  end;
{$ENDIF}
end; {SkipComment}

// Skip Over Leading White Space
procedure TPreProcessor.SkipWhite;
begin
SkipComment;
while Look in [' ', TAB, LF, CR] do
  begin
  GetChar;
  SkipComment;
  end;
end; {SkipWhite}

// Skip Over a Comma if it exists.
function TPreProcessor.SkipComma : boolean;
begin
result := Token = CSTI_Comma;
if result then
  GetToken;
end; {SkipComma}

procedure TPreProcessor.SkipToNewLine;
begin
while not IsNewLine(Look) and (Look <> #0) do
  GetChar;
end; {SkipToNewLine}

Procedure TPreProcessor.SkipNewLine;
begin
if look = CR then
  begin
  GetChar;
  if look = LF then
    GetChar;
  end;
end; {SkipNewLine}

// Get an Identifier
Procedure TPreProcessor.GetIdentifier;
var
  index : integer;
  StopLooping : boolean;
begin
fToken.Token := CSTI_Identifier;  
if not IsAlpha(Look) then
  Expected('Identifier',Look);
setLength(fToken.Value,8);
index := 0;
StopLooping := false;
repeat
  while IsAlNum(Look) do
    begin
    if look = HexadecimalChar then
      Scanner_Abort('Invalid char '''+HexadecimalChar+''' in identifier name');
    inc(index);
    if index >= length(fToken.Value) then
      setlength(fToken.Value,index+8);
    {$IFDEF CaseSensitive}
      fToken.Value[index] := Look;
    {$ELSE}
      fToken.Value[index] := UpCase(Look);
    {$ENDIF}
    GetChar;
    SkipComment;
    end;
  if (look = '-') and IsAlNum(NextLook) then
    begin
    GetChar;
    inc(index);
    if index >= length(fToken.Value) then
      setlength(fToken.Value,index+8);
    fToken.Value[index] := '-';  
    end
  else
    begin
    setLength(fToken.Value,index);
    StopLooping := true;
    end
until StopLooping;
end; {GetIdentifier}

// Get a Number
Procedure TPreProcessor.GetNumber;
var
  index : integer;
  IsFloat : boolean;
begin
fToken.Token := CSTI_Integer;
if not IsDigit(Look) then
  Expected('Integer or floatpoint',Look);
setLength(ftoken.Value,16);
index := 0;
if look = HexadecimalChar then
  ftoken.Token := CSTI_HexInt;
IsFloat := false;
while IsDigit(Look) or (look = '.') do
  begin
  inc(index);
  if index >= length(ftoken.Value) then
    setlength(ftoken.Value,index+16);
  if look = '.' then
    begin
    if not IsFloat then
      begin
      ftoken.Token := CSTI_Real;
      IsFloat := true;
      end
    else if ftoken.Token = CSTI_HexInt then
      Scanner_Abort('Invalid number format, mixing Hexadecimal with floatpoint')
    else
      Scanner_Abort('Invalid floatpoint number, duplicate ''.''')
    end
  else if look = HexadecimalChar then
    Scanner_Abort('Invalid Hexadecimal number, duplicate '''+HexadecimalChar+'''');
  ftoken.Value[index] := Look;
  GetChar;
  SkipComment;
  end;
if index <> 0 then
  begin
  if IsFloat and (ftoken.value[index] = '.') then
    begin // demote the variable type
    ftoken.Token := CSTI_Integer;
    dec(index);
    // emit a warning
//    Hint(ftoken.position,'Floatpoint demoted to integer : '''+ftoken.value+'.''');
    end;
  Setlength(ftoken.Value,index);
  end
else
  ftoken.Value := ''; 
end; {GetNumber}

// Get a string
procedure TPreProcessor.GetString;
var
  index : integer;
  tmp : string;
begin
fToken.Token := CSTI_String;
if (Look <> StringChar) then
  Expected('String character : '''+StringChar+'''',Look);
GetChar; // eat the start of string char;
setLength(tmp,8);
index := 0;
fToken.value := '';
Repeat
  // check to see if this is the end of the string or just imbedding a ' into the string
  // or it is the end of the string
  if (look = StringChar) then
    begin
    GetChar;
    if (look <> StringChar) then
      begin
      if index <> 0 then
        begin
        setLength(tmp,index);
        fToken.value := tmp;
        end;
      exit;
      end;
    GetChar;
    InsertString( NextCharIndex-1, StringChar );
    end;
  inc(index);
  if index > length(tmp) then
    setlength(tmp,index+8);
  tmp[index] := Look;  
  GetChar;
until IsNewLine(look) or (look = #0);
if IsNewLine(look) then
  Scanner_Abort('Unterminated string')
else if Look = #0 then // end of file
  Scanner_Abort('Unexpected end of file in string')
else
  Scanner_Abort('Error tokenizing string in lexical scanner');
end; {GetString}

Procedure TPreProcessor.GetOperator;
var
  OldPos : integer;
  OldLook : char;
  OldNextLook : char;
  NotValid : boolean;
begin
ftoken.Token := CSTINT_WhiteSpace;
if not IsOp(Look) then
  Expected('Operator',Look);
// biuld the operator type
ftoken.value := look;
case Look of
  // single char operators
  '+' : ftoken.Token := CSTI_Plus;
  '-' : ftoken.Token := CSTI_Minus;
  '*' : ftoken.Token := CSTI_Multiply;
  '/' : ftoken.Token := CSTI_Divide;
  ';' : ftoken.Token := CSTI_SemiColon;
  '(' : ftoken.Token := CSTI_OpenRound;
  ')' : ftoken.Token := CSTI_CloseRound;
  '.' : ftoken.Token := CSTI_Period;
  ',' : ftoken.Token := CSTI_Comma;
  ':' : ftoken.Token := CSTI_Colon;
  ']' : ftoken.Token := CSTI_CloseBlock;
  '{' : ftoken.Token := CSTII_Begin;
  '}' : ftoken.Token := CSTII_End;
  '~' : ftoken.Token := CSTI_LogicalNot;

end; {case}
if ftoken.Token <> CSTINT_WhiteSpace then
  begin // a token has been parsed
  GetChar;
  exit;
  end;
// go to the next posible token size
case Look of
  // multi sized operators
  '|' :
    begin
    if NextLook = '|' then
      begin
      ftoken.value := look + GetChar;
      ftoken.Token := CSTI_LogicalOr;
      end
    else
      ftoken.Token := CSTI_BitwiseOr
    end;
  '&' :
    begin
    if NextLook = '&' then
      begin
      ftoken.value := look + GetChar;
      ftoken.Token := CSTI_LogicalAnd;
      end
    else
      ftoken.Token := CSTI_BitwiseAnd;
    end;
  '^' :
    begin
    if NextLook = '^' then
      begin
      ftoken.value := look + GetChar;
      ftoken.Token := CSTI_LogicalXor
      end
    else
      ftoken.Token := CSTI_BitwiseXor;
    end;
  '!' :
    begin
    if NextLook = '=' then
      begin
      ftoken.value := look + GetChar;
      ftoken.Token := CSTI_NotEqual;
      end
    else
      ftoken.Token := CSTI_BitwiseNot;
    end;
  '=' :
    begin
    if NextLook = '=' then
      begin
      ftoken.value := look + GetChar;
      ftoken.Token := CSTI_Equal;
      end
    else
      ftoken.Token := CSTI_Assignment;
    end;
  '[' :
    begin 
    OldPos := NextCharIndex;
    OldLook := Look;
    OldNextLook := NextLook;
    NotValid := true;
    try
      GetChar;
      SkipWhite;
      if IsDigit(Look) or (look = '-') then
        begin
        if (look = '-') then
          begin
          GetChar;
          SkipWhite;
          GetNumber;
          ftoken.Value := '-'+ftoken.Value;
          end
        else GetNumber;
        case ftoken.Token of
          CSTI_Integer : ftoken.Value := IntToStr( Round( StrToInt( ftoken.Value ) * LinearConst ));
          CSTI_Real :
            begin
            ftoken.Value := IntToStr( Round( StrToFloat( ftoken.Value ) * LinearConst ));
            ftoken.Token := CSTI_Integer;
            end;
          else Scanner_Abort('Internal error in TPreProcessor.GetOperator(''['')');
        end;
        SkipWhite;
        NotValid := Look <> ']';
        end;
    except
    end;
    if NotValid then
      begin
      fNextCharIndex := OldPos;
      Look := OldLook;
      NextLook := OldNextLook;
      ftoken.Token := CSTI_OpenBlock;
      end;
    end;
  '<' :
    if NextLook = '=' then
      begin
      ftoken.value := look + GetChar;
      ftoken.Token := CSTI_LessEqual;
      end
    else
      begin
      OldPos := NextCharIndex;
      OldLook := Look;
      OldNextLook := NextLook;
      NotValid := true;
      try
        GetChar;      
        SkipWhite;
        if IsDigit(Look) or (look = '-') then
          begin
          if (look = '-') then
            begin
            GetChar;
            SkipWhite;
            GetNumber;
            ftoken.Value := '-'+ftoken.Value;
            end
          else GetNumber;
          case ftoken.Token of
            CSTI_Integer : ftoken.Value := IntToStr( Round( StrToInt( ftoken.Value ) * AngularConst ));
            CSTI_Real :
              begin
              ftoken.Value := IntToStr( Round( StrToFloat( ftoken.Value ) * AngularConst ));
              ftoken.Token := CSTI_Integer;
              end;
          else Scanner_Abort('Internal error in TPreProcessor.GetOperator(''<'')');
          end;
          SkipWhite;
          NotValid := Look <> '>';
          end;
      except
      end;
      if NotValid then
        begin
        fNextCharIndex := OldPos;
        Look := OldLook;
        NextLook := OldNextLook;
        ftoken.Token := CSTI_Less;
        end;
      end;
  '>' :
    if (NextLook = '=') then
      begin
      ftoken.value := look + GetChar;
      ftoken.Token := CSTI_GreaterEqual
      end
    else ftoken.Token := CSTI_Greater;
end; {case}
if ftoken.Token = CSTINT_WhiteSpace then
  Scanner_Abort('Unknown operator type : '''+ ftoken.value +'''');
GetChar;
end; {GetOperator}

Function TPreProcessor.TillNewLine( DoSkipComments : boolean = true ) : string;
var
  len, count : integer;
begin
len := 16;
Setlength( Result, len );
count := 0;
if DoSkipComments then
  SkipComment;
while not IsNewLine(Look) and (look <> #0) do
  begin
  inc( count );
  if count > len then
    begin
    len := count * 2;
    Setlength(Result, len);
    end;
  Result[Count] := Look;
  GetChar;
  if DoSkipComments then
    SkipComment;  
  end;
while (count > 0) and IsNewLine(Result[count]) do
  begin
  dec(count);
  NextCharIndex := NextCharIndex-1; 
  end;
setlength(Result,count);
end; {TillNewLine}

// Lexical Scanner
Procedure TPreProcessor.GetToken;
var
  NewText : string;
  Done,oldDoExpandMacro : boolean;
begin
// change the position of the current token
repeat
  SkipWhite;
  fToken.Position := NextCharIndex-1;
  repeat
    Done := true;
    SkipWhite;
    if IsAlpha(Look) then
      begin
      GetIdentifier;
      if Assigned(OnExpandMacro) and DoExpandMacro and
         OnExpandMacro(fToken.Value, NewText) then
       begin
       RemoveString( fToken.Position, NextCharIndex - fToken.Position-1, false );
       InsertString( NextCharIndex-1, NewText );
       Done := false;
       end;
      end
    else if IsDigit(Look) then
      GetNumber
    else if look = StringChar then
      GetString
    else if Look = PreProcessorCommand then
      begin
      RemoveString( NextCharIndex-1,1);
//      GetChar;
      GetIdentifier;
      if (fToken.value <> '') then
        begin
        If assigned(OnPreProcessorCommand) then
          begin
          oldDoExpandMacro := DoExpandMacro;
          try
            DoExpandMacro := false;
            OnPreProcessorCommand( self )
          finally
            DoExpandMacro := oldDoExpandMacro;
          end;
          end
        else
          begin
          SkipToNewLine;
          SkipNewLine;
          end;
        end;
      end
    else if IsOp(Look) then
      GetOperator
    else if Look = #0 then
      begin
      fToken.Token := CSTI_EOF;
      fToken.Value := '';
      end
    else
      begin
      fToken.Value := Look;
      fToken.Token := CSTI_Identifier;
      //  asm int 3 end; // Unknown token!!!
      GetChar;
      end;
  until Done;
until (fToken.Token <> CSTINT_WhiteSpace);
end; {GetToken}


Procedure TPreProcessor.Match( aToken : TToken );
var ErrorAddress : Pointer;
begin
if Token = aToken then
  GetToken
else
  begin
  asm
    MOV eax,[EBP+4];
    MOV ErrorAddress,eax;
  end;
  if Token = CSTI_Identifier then
    Expected(TokenName(aToken),Value,ErrorAddress)
  else
    Expected(TokenName(aToken),TokenName(Token),ErrorAddress);
  end;
end; {Match}

function TPreProcessor.IfMatch( aToken : TToken ): boolean;
begin
if Token = CSTI_EOF then
  Match( aToken );
result := Token = aToken;
if result then
  GetToken;
end; {IfMatch}

end.
